package com.bergerkiller.bukkit.nolagg;

import java.io.DataInputStream;
import java.io.IOException;
import java.lang.ref.WeakReference;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;

import org.bukkit.craftbukkit.CraftChunk;
import org.bukkit.craftbukkit.CraftWorld;

import com.bergerkiller.bukkit.common.Task;
import com.bergerkiller.bukkit.nolagg.bufferedtag.NBTBaseBuffered;
import com.bergerkiller.bukkit.nolagg.datapool.Pool;
import com.bergerkiller.bukkit.nolagg.datapool.PoolBuffer;
import com.bergerkiller.bukkit.nolagg.datapool.PoolData;

import net.minecraft.server.Chunk;
import net.minecraft.server.ChunkProviderServer;
import net.minecraft.server.Entity;
import net.minecraft.server.EntityTypes;
import net.minecraft.server.NBTTagCompound;
import net.minecraft.server.NBTTagList;
import net.minecraft.server.NibbleArray;
import net.minecraft.server.TileEntity;
import net.minecraft.server.World;
import net.minecraft.server.WorldServer;

public class BufferedChunk extends Chunk implements PoolData<ChunkBuffer> {
	private static Field weakReference = null;
	private static Map<ChunkCoord, BufferedChunk> unloadedChunks = new HashMap<ChunkCoord, BufferedChunk>();
	static {
		try {
			weakReference = CraftChunk.class.getDeclaredField("weakChunk");
			weakReference.setAccessible(true);
		} catch (Throwable t) {
			NoLagg.log(Level.SEVERE, "Failed to initialize craft chunk weak reference removal: weird glitches may start occurring!");
		}
	}
	
	private static Pool<ChunkBuffer> pool = new Pool<ChunkBuffer>(100);
	public static String getBufferSizeName() {
		int size = pool.getDataSize() / 1000;
		if (size < 1000) {
			return size + " kb";
		} else {
			return Util.round((double) size / 1000, 1) + " mb";
		}
	}
	
	private static class ChunkCoord {
		public ChunkCoord(final World world, final int x, final int z) {
			this.world = world;
			this.x = x;
			this.z = z;
		}
		public final World world;
		public final int x;
		public final int z;
		public int hashCode() {
			int hash = 3;
	        hash = 21 * hash + this.world.hashCode();
	        hash = 21 * hash + (this.x ^ (this.x >> 16));
	        hash = 21 * hash + (this.z ^ (this.z >> 16));
	        return hash;
		}
		public boolean equals(Object o) {
			if (o == this) return true;
			if (o instanceof ChunkCoord) {
				ChunkCoord coord = (ChunkCoord) o;
				return coord.x == this.x && coord.z == this.z && coord.world == this.world;
			}
			return false;
		}
	}
	
	public static BufferedChunk removeUnloaded(final World world, final int x, final int z) {
		synchronized (unloadedChunks) {
			return unloadedChunks.remove(new ChunkCoord(world, x, z));
		}
	}
	
	public void markUnloaded() {
		synchronized (unloadedChunks) {
			unloadedChunks.put(new ChunkCoord(this.world, this.x, this.z), this);
		}
	}
		
	public BufferedChunk(final World world, final int x, final int z) {
		super(world, x, z);
		this.buffer = pool.create(this);
		this.buffer.getData().apply(this);
	}
	
	public boolean isUnloaded = false;
	
	private PoolBuffer<ChunkBuffer> buffer;
	@Override
	public ChunkBuffer getNewData() {
		return new ChunkBuffer();
	}
	
	public boolean isFree() {
		return this.buffer == null;
	}
		
	public void loadNOP() {
		if (this.isFree()) {
			//added back in the cycle somehow...
			this.buffer = pool.create(this);
			this.buffer.getData().apply(this);
			removeUnloaded(this.world, this.x, this.z);
			//generate bukkit chunk
	        // CraftBukkit start
            org.bukkit.craftbukkit.CraftWorld cworld = this.world.getWorld();
            this.bukkitChunk = (cworld == null) ? null : cworld.popPreservedChunk(this.x, this.z);
            if (this.bukkitChunk == null) {
                this.bukkitChunk = new org.bukkit.craftbukkit.CraftChunk(this);
            }
            //make sure we load the newly created data
            BufferedChunkRegionLoader loader = BufferedChunkRegionLoader.convert((WorldServer) this.world);
            if (loader == null) {
            	NoLagg.log(Level.SEVERE, "The buffered chunk " + this.toString() + " ended up unloaded without a buffered loader ready to load it again!");
            } else {
            	DataInputStream stream = loader.getChunkStream(this.x, this.z);
            	if (stream == null) {
            		NoLagg.log(Level.SEVERE, "The buffered chunk " + this.toString() + " ended up unloaded without a source file to load it again!");
            	} else {
            		try {
                		this.load(loader, stream);
            		} catch (Throwable t) {
            			NoLagg.log(Level.SEVERE, "The buffered chunk " + this.toString() + " ended up unloaded and failed to load again!");
            			t.printStackTrace();
            		}
            	}
            }
		}
		super.loadNOP();
	}
		
	@SuppressWarnings("unchecked")
	public void setFree() {
		if (this.buffer == null) return;
		//make sure this occurrence is synchronized!
		final BufferedChunk chunk = this;
		new Task(NoLagg.plugin) { 
			public void run() {
				if (chunk.buffer != null && chunk.isUnloaded()) {
					chunk.buffer.setFree(this);
					chunk.buffer = null;
					chunk.b = null;
					chunk.g = chunk.h = chunk.i = null;
					if (chunk.bukkitChunk != null && weakReference != null) {
						//break link
						try {
							WeakReference<Chunk> ref = (WeakReference<Chunk>) weakReference.get(chunk.bukkitChunk);
							ref.clear();
							chunk.bukkitChunk = null;
						} catch (Throwable t) {
							t.printStackTrace();
							weakReference = null;
						}
					}
				}
			}
		}.start(1);
	}
			
	public void finalize() throws Throwable {
		if (this.buffer != null) {
			this.buffer.dispose();
			this.buffer = null;
		}
		super.finalize();
	}
		
	public ChunkProviderServer getChunkProviderServer() {
		return ((WorldServer) this.world).chunkProviderServer;
	}
		
	public boolean isUnloaded() {
		return this.getChunkProviderServer().chunks.get(this.x, this.z) != this;
	}
	
	public String toString() {
		return "[" + this.world.getWorld().getName() + ", " + this.x + ", " + this.z + "]";
	}
	
	public CraftWorld getWorld() {
		return this.world.getWorld();
	}
	
	/*
	 * Fix to prevent index out of bounds for 'invalid' tile entities
	 */
	public void a(int i, int j, int k, TileEntity tileentity) {
		if (i < 0 || i > 15 || j < 0 || j > 127 || k < 0 || k > 15) {
			NoLagg.log(Level.WARNING, "Tile entity got stored wrongly, chunk " + this.toString() + " could be corrupted!");
			return;
		}
		super.a(i, j, k, tileentity);
	}
	
	public boolean containsCoordinates(int x, int z) {
		return x >= 0 && x < 16 && z >= 0 && z < 16;
	}
	
	/*
	 * Prevents out of bounds exceptions for chunks
	 */
	public int getTypeId(int x, int y, int z) {
		if (y > 127 || y < 0) return 0; //return air above/under chunk height
		if (!containsCoordinates(x, z)) {
			NoLagg.log(Level.WARNING, "Block type id was requested outside chunk bounds: [" + this.world.getWorld().getName() + ", " + (x + this.x << 4) + ", " + y + ", " + (z | this.z << 4) + "]");
			Thread.dumpStack();
			return 0;
		}
		return super.getTypeId(x, y, z);
	}
	public int getData(int x, int y, int z) {
		if (y > 127 || y < 0) return 0; //return 0 above/under chunk height
		if (!containsCoordinates(x, z)) {
			NoLagg.log(Level.WARNING, "Block type id was requested outside chunk bounds: [" + this.world.getWorld().getName() + ", " + (x + this.x << 4) + ", " + y + ", " + (z | this.z << 4) + "]");
			Thread.dumpStack();
			return 0;
		}
		return super.getData(x, y, z);
	}
		
	public NBTTagCompound getBufferedTag(DataInputStream stream) throws Throwable {
		byte type = stream.readByte();
		if (type != 10) throw new IOException("Root tag must be a named compound tag");
		String name = stream.readUTF();
		//BufferedTag reading with buffers
		final Map<String, byte[]> buffers = new HashMap<String, byte[]>(5);
		buffers.put("Blocks", this.b);
		buffers.put("Data", this.g.a);
		buffers.put("SkyLight", this.h.a);
		buffers.put("BlockLight", this.i.a);
		buffers.put("HeightMap", this.heightMap);
		return (NBTTagCompound) NBTBaseBuffered.readFrom(type, name, buffers, stream);
	}
	public void load(BufferedChunkRegionLoader loader, NBTTagCompound comp) throws Throwable {
		this.load(loader, comp, comp.getCompound("Level"));
	}
	public void load(BufferedChunkRegionLoader loader, DataInputStream stream) throws Throwable {
		this.load(loader, this.getBufferedTag(stream));
	}
	
	public void load(BufferedChunkRegionLoader loader, NBTTagCompound comp, NBTTagCompound level) throws Throwable {
		if (!level.hasKey("xPos")) level.setInt("xPos", this.x);
		if (!level.hasKey("zPos")) level.setInt("zPos", this.z);
		//do the coordinates match?
		int realX = level.getInt("xPos");
		int realZ = level.getInt("zPos");
		if (realX != this.x || realZ != this.z) {
			loader.relocate(this, comp, realX, realZ);
			throw new WrongChunkCoordinatesException();
		}
		
		
		this.b = level.getByteArray("Blocks");
		this.g = Util.getNibble(this.g, level.getByteArray("Data"), this.world.heightBits);
		this.h = Util.getNibble(this.h, level.getByteArray("SkyLight"), this.world.heightBits);
		this.i = Util.getNibble(this.i, level.getByteArray("BlockLight"), this.world.heightBits);
		
		this.heightMap = level.getByteArray("HeightMap");
		this.done = level.getBoolean("TerrainPopulated");
		if (!this.g.a()) {
			this.g = new NibbleArray(this.b.length, this.world.heightBits);
		}

		if (this.heightMap == null || !this.h.a()) {
			this.heightMap = new byte[256];
			this.h = new NibbleArray(this.b.length, this.world.heightBits);
			this.initLighting();
		}

		if (!this.i.a()) {
			this.i = new NibbleArray(this.b.length, this.world.heightBits);
			this.a();
		}

		NBTTagList nbttaglist = level.getList("Entities");

		if (nbttaglist != null) {
			for (int k = 0; k < nbttaglist.size(); ++k) {
				NBTTagCompound nbttagcompound1 = (NBTTagCompound) nbttaglist.get(k);
				Entity entity = EntityTypes.a(nbttagcompound1, this.world);

				this.s = true;
				if (entity != null) {
					this.a(entity);
				}
			}
		}

		NBTTagList nbttaglist1 = level.getList("TileEntities");

		if (nbttaglist1 != null) {
			for (int l = 0; l < nbttaglist1.size(); ++l) {
				NBTTagCompound nbttagcompound2 = (NBTTagCompound) nbttaglist1.get(l);
				TileEntity tileentity = TileEntity.c(nbttagcompound2);

				if (tileentity != null) {
					this.a(tileentity);
				}
			}
		}

		if (level.hasKey("TileTicks")) {
			NBTTagList nbttaglist2 = level.getList("TileTicks");

			if (nbttaglist2 != null) {
				for (int i1 = 0; i1 < nbttaglist2.size(); ++i1) {
					NBTTagCompound nbttagcompound3 = (NBTTagCompound) nbttaglist2.get(i1);

					this.world.d(nbttagcompound3.getInt("x"), nbttagcompound3.getInt("y"), nbttagcompound3.getInt("z"), nbttagcompound3.getInt("i"), nbttagcompound3.getInt("t"));
				}
			}
		}
	}

}