package com.bergerkiller.bukkit.nolagg;

import java.util.logging.Level;

import org.bukkit.craftbukkit.CraftWorld;

import com.bergerkiller.bukkit.nolagg.sending.ChunkSendQueue;

import net.minecraft.server.Block;
import net.minecraft.server.Chunk;
import net.minecraft.server.NibbleArray;
import net.minecraft.server.WorldServer;

public class ChunkFixLightingOperation implements ThreadOperation {

	public static int processed = 0;
	
	private final int x, z;
	private final WorldServer world;
	private NibbleArray lightdata;
	private byte[] typedata;
	private byte[] heightdata;
	public ChunkFixLightingOperation(org.bukkit.Chunk chunk) {
		this(ChunkHandler.getNative(chunk));
	}
	public ChunkFixLightingOperation(Chunk chunk) {
		this.world = (WorldServer) chunk.world;
		this.x = chunk.x;
		this.z = chunk.z;
	}
	
	@Override
	public void start() {
		Chunk chunk = this.getChunk(this.x, this.z);
		if (chunk != null) {
			this.lightdata = new NibbleArray(new byte[chunk.h.a.length], this.world.heightBits);
			this.typedata = new byte[chunk.b.length];
			this.heightdata = new byte[chunk.heightMap.length];
			System.arraycopy(chunk.h.a, 0, this.lightdata.a, 0, this.lightdata.a.length);
			System.arraycopy(chunk.b, 0, this.typedata, 0, this.typedata.length);
			System.arraycopy(chunk.heightMap, 0, this.heightdata, 0, this.heightdata.length);
		}
	}
	
	private int getHeight(final int x, final int z) {
		return this.heightdata[x << 4 | z] & 255;
	}
	private Chunk getChunk(final int x, final int z) {
		return this.world.chunkProviderServer.chunks.get(x, z);
	}
	private int getTypeId(final int x, final int y, final int z) {
		return this.typedata[x << this.world.heightBitsPlusFour | y << this.world.heightBits | z] & 255;
	}
	private int getLightLevel(int x, final int y, int z) {
		if (y <= 0 || y >= 127) return 0;
		if (x >= 0 && z >= 0 && x < 16 && z < 16) {
			return this.lightdata.a(x, y, z);
		}
		
	    Chunk chunk = this.getChunk(this.x + (x >> 4), this.z + (z >> 4));
		if (chunk == null) return 0;
		x -= (chunk.x - this.x) << 4;
		z -= (chunk.z - this.z) << 4;
		return chunk.h.a(x, y, z);
	}
	private boolean fixBlockLighting(final int x, final int y, final int z) {
		final int typeid = this.getTypeId(x, y, z);
		if (!Block.o[typeid]) {
			final int factor = Math.max(1, Block.lightBlock[typeid]);
			final int light = this.getLightLevel(x, y, z);
			//actual editing here
			int newlight = light + factor;
			newlight = Math.max(newlight, getLightLevel(x - 1, y, z));
			newlight = Math.max(newlight, getLightLevel(x + 1, y, z));
			newlight = Math.max(newlight, getLightLevel(x, y, z - 1));
			newlight = Math.max(newlight, getLightLevel(x, y, z + 1));
			newlight = Math.max(newlight, getLightLevel(x, y - 1, z));
			newlight = Math.max(newlight, getLightLevel(x, y + 1, z));
			newlight -= factor;
			//when exceeding in loops, just pick the highest value
			if (newlight > light) {
				this.lightdata.a(x, y, z, newlight);
				return true;
			}
		}
		return false;
	}
	
	public String toString() {
		CraftWorld world = this.world.getWorld();
		if (world == null) {
			return "[" + this.x + "/" + this.z + "]";
		} else {
			return "[" + this.x + "/" + this.z + "/" + world.getName() + "]";
		}
	}

	@Override
	public void run() {
		if (this.lightdata == null) return;
		boolean haserror = true;
		int loops = -1;
		int x, y, z;
		while (haserror) {
			if (loops > 20) {
				NoLagg.log(Level.WARNING, "Failed to fix all lighting issues in chunk " + this.toString());
				break;
			}
			haserror = false;
			loops++;
			for (x = 0; x < 16; x++) {
				for (z = 0; z < 16; z++) {
					for (y = this.getHeight(x, z); y > 0; --y) {
						if (fixBlockLighting(x, y, z)) {
							haserror = true;
						}
					}
				}
			}
		}
	}

	@Override
	public void end() {
		if (this.lightdata == null) return;
		//transfer block data
		Chunk chunk = this.getChunk(this.x, this.z);
		if (chunk != null) { 
			System.arraycopy(this.lightdata.a, 0, chunk.h.a, 0, this.lightdata.a.length);
			ChunkSendQueue.queueGlobal(chunk);
		}
		processed++;
	}

}
